# SEIR as a compartmented model
#
# Copyright (C) 2017--2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import CompartmentedModel, Node, Edge
import sys
if sys.version_info >= (3, 8):
    from typing import Final, Dict, Any
else:
    # backport compatibility with older typing
    from typing import Dict, Any
    from typing_extensions import Final

class SEIR(CompartmentedModel):
    '''The Susceptible-Exposed-Infected-Removed :term:`compartmented model of disease`.
    A susceptible node becomes exposed when infected by either an exposed or an infected
    neighbour. Exposed nodes become infected (symptomatic) to infected and then recover
    to removed.
    
    In contrast to the more familiar :class:`SIR` model, SEIR has two
    compartments that can pass infection. The utility of the model from two aspects:
    exploring what fraction of the contract tree arises from infected individuals
    *versus* exposed individuals, capturing the significance of asymptomatic infection;
    and allowing countermeasures to be applied to symptomatic individuals whose presence
    could be more easily detected than those who are exposed by asymptomatic.

    The SEIR model in `epydemic` is very flexible, allowing different infection probabilities
    for susceptible-exposed or susceptible-infected interactions. The initial seed population
    is placed into :attr:`EXPOSED`, rather than into :attr:`INFECTED` as happens
    for :class:`SIR`.'''
    
    # Model parameters
    P_EXPOSED : Final[str] = 'epydemic.SEIR.pExposed'              #: Parameter for probability of initially being exposed.
    P_INFECT_ASYMPTOMATIC : Final[str] = 'epydemic.SEIR.pInfectA'  #: Parameter for probability of infection on contact with an exposed individual
    P_INFECT_SYMPTOMATIC : Final[str] = 'epydemic.SEIR.pInfect'    #: Parameter for probability of infection on contact with a symptomatic individual.
    P_SYMPTOMS : Final[str] = 'epydemic.SEIR.pSymptoms'            #: Parameter for probability of becoming symptomatic after exposure. 
    P_REMOVE : Final[str] = 'epydemic.SEIR.pRemove'                #: Parameter for probability of removal (recovery).
    
    # Possible dynamics states of a node for SIR dynamics
    SUSCEPTIBLE : Final[str] = 'epydemic.SEIR.S'        #: Compartment for nodes susceptible to infection.
    EXPOSED : Final[str] = 'epydemic.SEIR.E'            #: Compartment for nodes exposed and infectious.
    INFECTED : Final[str] = 'epydemic.SEIR.I'           #: Compartment for nodes symptomatic and infectious.
    REMOVED : Final[str] = 'epydemic.SEIR.R'            #: Compartment for nodes recovered/removed.

    # Loci containing the edges at which dynamics can occur
    SE : Final[str] = 'epydemic.SEIR.SE'                #: Edge able to transmit infection from an exposed individual.
    SI : Final[str] = 'epydemic.SEIR.SI'                #: Edge able to transmit infection from an infected individual.

    def __init__(self):
        super(SEIR, self).__init__()

    def build(self, params : Dict[str, Any]):
        '''Build the SEIR model.

        :param params: the model parameters'''
        super(SEIR, self).build(params)

        pExposed = params[self.P_EXPOSED]
        pInfectA = params[self.P_INFECT_ASYMPTOMATIC]
        pInfect = params[self.P_INFECT_SYMPTOMATIC]
        pSymptoms = params[self.P_SYMPTOMS]
        pRemove = params[self.P_REMOVE]

        self.addCompartment(self.SUSCEPTIBLE, 1.0 - pExposed)
        self.addCompartment(self.EXPOSED, pExposed)
        self.addCompartment(self.INFECTED, 0.0)
        self.addCompartment(self.REMOVED, 0.0)

        self.trackEdgesBetweenCompartments(self.SUSCEPTIBLE, self.EXPOSED, name=self.SE)
        self.trackEdgesBetweenCompartments(self.SUSCEPTIBLE, self.INFECTED, name=self.SI)
        self.trackNodesInCompartment(self.EXPOSED)
        self.trackNodesInCompartment(self.INFECTED)

        self.addEventPerElement(self.SE, pInfectA, self.infectAsymptomatic)
        self.addEventPerElement(self.SI, pInfect, self.infect)
        self.addEventPerElement(self.EXPOSED, pSymptoms, self.symptoms)
        self.addEventPerElement(self.INFECTED, pRemove, self.remove)

    def infectAsymptomatic(self, t : float, e : Node):
        '''Perform an infection event when an :attr:`EXPOSED` individual infects
        a neighbouring :attr:`SUSCEPTIBLE`, rendering them :attr:`EXPOSED` in turn.  
        The default calls :meth:`infect` so that infections by way of exposed or
        symptomatic individuals are treated in the same way. Sub-classes can override this
        to, for example, record that the infection was passed asymptomatically.

        :param t: the simulation time
        :param e: the edge transmitting the infection'''
        self.infect(t, e)

    def infect(self, t : float, e : Edge):
        '''Perform an infection event when an :attr:`INFECTED` individual infects
        a neighbouring :attr:`SUSCEPTIBLE`, rendering them :attr:`EXPOSED` in turn.  
        
        :param t: the simulation time
        :param e: the edge transmitting the infection'''
        (n, _) = e
        self.changeCompartment(n, self.EXPOSED)
        self.markOccupied(e, t)

    def symptoms(self, t : float, n : Node):
        '''Perform the symptoms-developing event. This changes the compartment of
        the node to :attr:`INFECTED`.

        :param t: the simulation time (unused)
        :param n: the node'''
        self.changeCompartment(n, self.INFECTED)

    def remove(self, t : float, n : Node):
        '''Perform a removal event. This changes the compartment of
        the node to :attr:`REMOVED`.

        :param t: the simulation time (unused)
        :param n: the node'''
        self.changeCompartment(n, self.REMOVED)
    
                
   



# A process to collect global network statistics
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import sys
from epydemic import Process
from networkx import degree_histogram, connected_components
if sys.version_info >= (3, 8):
    from typing import Final, Dict, Any
else:
    # backport compatibility with older typing
    from typing import Dict, Any
    from typing_extensions import Final


class NetworkStatistics(Process):
    '''A process that collects statistics about the final network. This
    process defines no events: it simply interrogates the underlying
    network and extracts a collection of descriptive statistics into
    the final experimental results. These are statistics commonly
    collected in different scenarios, which can thus be provided as
    standard.

    '''

    # Experimental results
    N : Final[str] = 'epydemic.networkstatistics.N'                    #: Result holding the order of the network.
    M : Final[str] = 'epydemic.networkstatistics.M'                    #: Result holding the total number of edges in the network.
    KMEAN : Final[str] = 'epydemic.networkstatistics.kmean'            #: Result holding the mean degree of nodes in the network.
    KDIST : Final[str] = 'epydemic.networkstatistics.k_distribution'   #: Result holding the degree histogram as an array.
    COMPONENTS : Final[str] = 'epydemic.networkstatistics.components'  #: Result holding the number of connected components in the network.
    LCC : Final[str] = 'epydemic.networkstatistics.lcc'                #: Result holding the size of the largest (giant) component
    SLCC : Final[str] = 'epydemic.networkstatistics.slcc'              #: Result holding the size of the second-largest component

    def __init__(self):
        super().__init__()

    def results(self) -> Dict[str, Any]:
        '''Extract the network summary statistics into a dict.

        :returns: a dict of experimental results'''
        res = super().results()

        # order statistics
        g = self.network()
        res[self.N] = g.order()
        res[self.M] = len(g.edges)

        # degree statistics
        hist = degree_histogram(g)
        ktotal = 0
        for i in range(len(hist)):
            ktotal += i * hist[i]
        res[self.KMEAN] = ktotal / g.order()
        res[self.KDIST] = hist

        # component statistics
        ccs = sorted(list(map(len, connected_components(g))), reverse=True)
        nccs = len(ccs)
        res[self.COMPONENTS] = nccs
        res[self.LCC] = ccs[0] if nccs > 0 else 0
        res[self.SLCC] = ccs[1] if nccs > 1 else 0

        return res



# A generator for powerlaw-with-cutoff networks
#
# Copyright (C) 2017--2021 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import NetworkGenerator
from networkx import Graph, configuration_model
from math import exp
from mpmath import polylog
import numpy
import sys
if sys.version_info >= (3, 8):
    from typing import Any, Dict, Optional, Callable, Final
else:
    # backport compatibility with older typing
    from typing import Any, Dict, Optional, Callable
    from typing_extensions import Final


class PLCNetwork(NetworkGenerator):
    '''A generator for networks with a powerlaw-with-cutoff degree distribution.

    PLC networks are commonly used to represent the basic structure of human contact
    networks, where we want to cutoff the degree to stop it becoming "too high" to
    be credible: :ref:`Newman <New02>` (and other works) use this topology extensively.
    It is characterised by two parameters, the exponent of the power law and the
    cutoff maximum degree after which the probability of finding nodes with larger degrees
    falls off exponentially. 

    :param params: (optional) experimental parameters
    :param limit: (optional) maximum number of independent networks to generate'''

    # Experimental parameters
    N : Final[str] = 'epydemic.generators.PLCNetwork.N'                 #: Experimental parameter for the order of the network.
    EXPONENT : Final[str] = 'epydemic.generators.PLCNetwork.exponent'   #: Experimental parameter for the exponent of the distribution.
    CUTOFF : Final[str] = 'epydemic.generators.PLCNetwork.cutoff'       #: Experimental parameter for the cutoff of the distribution.

    def __init__(self, params : Dict[str, Any] =None, limit : Optional[int] =None):
        super(PLCNetwork, self).__init__(params, limit)

    def _makePowerlawWithCutoff(self, alpha : float, kappa : float) -> Callable[[int], float]:
        '''Create a model function for a powerlaw distribution with exponential cutoff.

        :param alpha: the exponent of the distribution
        :param kappa: the degree cutoff
        :returns: a model function'''
        C = polylog(alpha, exp(-1.0 / kappa))
        def p( k : int ) -> float:
            return (pow((k + 0.0), -alpha) * exp(-(k + 0.0) / kappa)) / C
        return p

    def _generateFrom(self, N : int, p : Callable[[int], float], maxdeg : int =100):
        '''Generate a random graph with degree distribution described
        by a model function.

        :param N: number of numbers to generate
        :param p: model function
        :param maxdeg: maximum node degree we'll consider (defaults to 100)
        :returns: a network with the given degree distribution'''
        rng = numpy.random.default_rng()
        ns = []
        t = 0
        for i in range(N):
            while True:
                # draw a random degree
                k = rng.integers(1, maxdeg)
            
                # do we include a node with this degree?
                if rng.random() < p(k):
                    # yes, add it to the sequence; otherwise, draw again
                    ns.append(k)
                    t += k
                    break

        # the final sequence of degrees has to sum to an even
        # number, as each edge has two endpoints
        # if the sequence is odd, remove an element and draw
        # another from the distribution, repeating until the
        # overall sequence is even
        while t % 2 != 0:
            # pick a node at random
            i = rng.integers(0, len(ns) - 1)

            # remove it from the sequence and from the total
            t -= ns[i]
            del ns[i]
            
            # choose a new node to replace the one we removed
            while True:
                # draw a new degree from the distribution
                k = rng.integers(1, maxdeg)
            
                # do we include a node with this degree?
                if rng.random() < p(k):
                    # yes, add it to the sequence; otherwise, draw again
                    ns.append(k)
                    t += k
                    break

        # populate the network using the configuration
        # model with the given degree distribution
        g = configuration_model(ns, create_using=Graph())
        return g

    def _generate(self, params : Dict[str, Any]) -> Graph:
        '''Generate the human contact network.

        :param params: the experimental parameters
        :returns: a network'''
        N = params[self.N]
        alpha = params[self.EXPONENT]
        kappa = params[self.CUTOFF]
        return self._generateFrom(N, self._makePowerlawWithCutoff(alpha, kappa))




# Test bitstreams
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import unittest

class BitstreamTest(unittest.TestCase):

    # ---------- Bitstream tests ----------

    def testBits(self):
        '''Test bitstreams carry on regardless.'''
        bs = Bitstream()
        bss = iter(bs)
        bits = []
        for _ in range(100000):
            bits.append(next(bss))
        ones = len([b for b in bits if b == 1])
        zeros = len([b for b in bits if b == 0])
        self.assertEqual(ones + zeros, len(bits))
        self.assertAlmostEqual(ones, zeros, delta=int(len(bits) * 0.05))


if __name__ == '__main__':
    unittest.main()



# Test standard network generators
#
# Copyright (C) 2017--2021 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Process, Dynamics,FixedNetwork, ERNetwork, BANetwork, PLCNetwork
import unittest
import networkx

class GeneratorTest(unittest.TestCase):
    
    def assertGraphsEqual( self, g1, g2 ):
        '''Test two graphs for equality.
        sd: should check for attributes as well?

        :param g1: the first graph
        :param g2: the second graph'''
        g1ns = set(g1.nodes())
        g2ns = set(g2.nodes())
        self.assertCountEqual(g1ns, g2ns)
        g1es = set(g1.edges())
        g2es = set(g2.edges())
        self.assertCountEqual(g1es, g2es)
        

    def testFixedNetwork( self ):
        '''Test we always return a fixed network.'''
        g = networkx.diamond_graph()   # doesn't matter what it as, as long as it's fixed
        gen = FixedNetwork(g)

        self.assertGraphsEqual(g, gen.generate())
        
        for _ in range(5):   # make sure the iterator is stable
            gp = gen.__next__()
            self.assertGraphsEqual(g, gp)

    def testLimit(self):
        '''Test we can limit the number of instances generated.'''
        g = networkx.diamond_graph()
        gen = FixedNetwork(g, limit=10)
        
        n = 0
        for _ in gen:
            n += 1
        self.assertEqual(n, 10)

    def testFixedNetworkGenerator(self):
        '''Test that adding a network creates the right generator.'''
        g = networkx.diamond_graph()
        p = Process()
        dyn = Dynamics(p)

        dyn.setNetworkGenerator(g)
        self.assertIsInstance(dyn.networkGenerator(), FixedNetwork)
        self.assertGraphsEqual(g, dyn.networkGenerator().generate())
        
    def testER(self):
        '''Test we can generate ER networks with all parameter combinations.'''
        param = dict()
        param[ERNetwork.N] = 1000

        # test using <k>
        param[ERNetwork.KMEAN] = 20
        gen = ERNetwork(param)
        _ = gen.generate()

        # test using phi
        del param[ERNetwork.KMEAN]
        param[ERNetwork.PHI] = 0.02
        gen = ERNetwork(param)
        g_= gen.generate()

        # test working with both
        param[ERNetwork.KMEAN] = 20
        gen = ERNetwork(param)
        _ = gen.generate()
       
        # test failing with neither
        del param[ERNetwork.KMEAN]
        del param[ERNetwork.PHI]
        gen = ERNetwork(param)
        with self.assertRaises(AttributeError):
            _ = gen.generate()

    def testBA(self):
        '''Test we can generate BA networks.'''
        param = dict()
        param[BANetwork.N] = 1000
        param[BANetwork.M] = 20
        gen = BANetwork(param)
        _ = gen.generate()
        
    def testPLC(self):
        '''Test we can  generate PLC networks.'''
        param = dict()
        param[PLCNetwork.N] = 1000
        param[PLCNetwork.EXPONENT] = 2
        param[PLCNetwork.CUTOFF] = 10
        gen = PLCNetwork(param)
        _ = gen.generate()

if __name__ == '__main__':
    unittest.main()



# Loci for dynamics epidemic model
#
# Copyright (C) 2017--2020 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Element, DrawSet
from networkx import Graph
from typing import Iterator


class Locus(DrawSet):
    '''The locus of dynamics, where changes can happen. Loci
    are filled with nodes or edges from a network and represent some (possibly
    complete) sub-set of the elements, typically being populated and re-populated as
    a process over the network evolves.

    A locus is basically a set with some additional methods to allow
    or more complex behaviours, including customisable drawing random
    selection of elements. The underlying set is implemented as a
    :class:`DrawSet` to ensure scalability, especially of the
    :meth:`draw` method whose performance is critical.

    :param name: the locus name

    '''

    def __init__(self, name: str):
        super().__init__()
        self._name: str = name

    def name(self) -> str:
        '''Returns the name of the locus.

        :returns: the locus' name'''
        return self._name


    # ---------- Handlers ----------

    def addHandler(self, g: Graph, e: Element):
        '''Handler called when an element is added to the network. This is used to indicate
        the the population of nodes or edges has changed.

        :param g: the network
        :param e: the element'''
        self.add(e)

    def leaveHandler(self, g: Graph, e: Element):
        '''Handler for when an element leaves the locus due to changes in circumstances,
        not changes in population.

        :param g: the network
        :param e: the element'''
        self.discard(e)

    def enterHandler(self, g: Graph, e: Element):
        '''Handler for when an element enters the locus due to changes in circumstances,
        not changes in population.

        :param g: the network
        :param e: the element'''
        self.add(e)

    def removeHandler(self, g: Graph, e: Element):
        '''Handler called when an element is removed from the network.  This is used to indicate
        the the population of nodes or edges has changed.

        :param g: the network
        :param e: the element'''
        self.discard(e)



# Gillespie dynamics base class
#
# Copyright (C) 2017--2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import math
import numpy                     # type: ignore
from typing import Dict, Any, Union
from networkx import Graph
from epydemic import Dynamics, Process, NetworkGenerator


class StochasticDynamics(Dynamics):
    '''A dynamics that runs stochastically in :term:`continuous time`. This is a
    very efficient and statistically exact approach, but requires that the
    statistical properties of the events making up the process are known. See
    :ref:`Gillespie 1976 <Gil76>` and :ref:`Gillespie 1977 <Gil77>` for a discussion
    of the technique.


    :param p: the process to run
    :param g: network or network generator (optional, can be provided later)'''

    def __init__(self, p : Process, g : Union[Graph, NetworkGenerator] =None):
        super().__init__(p, g)

    def do(self, params : Dict[str, Any]) -> Dict[str, Any]:
        '''Run the simulation using Gillespie dynamics.

        :param params: the experimental parameters
        :returns: the experimental results dict'''

        # run the dynamics
        proc = self.process()
        rng = numpy.random.default_rng()
        t = 0
        events = 0
        while not proc.atEquilibrium(t):
            # pull the transition dynamics at this timestep
            transitions = self.eventRateDistribution(t)

            # compute the total rate of transitions for the entire network
            a = 0.0
            for (_, r, _) in transitions:
                a = a + r
            if a == 0.0:
                break              # no events with non-zero rates

            # shuffle the transitions
            #random.shuffle(transitions)

            # calculate the timestep delta
            r1 = rng.random()
            dt = (1.0 / a) * math.log(1.0 / r1)

            # calculate which event happens
            (l, _, ef) = transitions[0]
            if len(transitions) > 1:
                # choose the rate threshold
                r2 = rng.random()
                xc = r2 * a

                # find the largest event for which the cumulative rates
                # are less than the random threshold
                xs = 0
                for v in range(0, len(transitions)):
                    (l, xsp, ef) = transitions[v]
                    if (xs + xsp) > xc:
                        break
                    else:
                        xs = xs + xsp

            # increment the time
            t = t + dt
            self.setCurrentSimulationTime(t)

            # fire any events posted for at or before this time
            events = events + self.runPendingEvents(t)

            # it's possible that posted events have removed all elements
            # from the chosen locus, in which case we simply continue
            # with the next event selection
            # sd: is this correct? or does it mess up the statistics too much?
            if len(l) > 0:
                # draw a random element from the chosen locus
                e = l.draw()

                # perform the event by calling the event function,
                # passing the dynamics, event time, network, and element
                ef(t, e)

                # increment the event counter
                events = events + 1

        # when we get here there may still be posted events that haven't
        # been run, and these are ignored: equilibrium overrides posting

        # add some more metadata
        (self.metadata())[self.TIME] = t
        (self.metadata())[self.EVENTS] = events

        # report results
        rc = self.experimentalResults()
        return rc



# A set with an efficient random draw, wrapping a balanced binary tree
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Element, TreeNode
from typing import Iterator


class DrawSet():
    '''A set re-implementation.

    This class provides a re-implementation of sets that also provides
    an efficient (:math:`O(\log N)`) method for drawing al element at random.

    We implement only tyhose parts of the set API that we need: perhaps
    ought to add the rest, for future-proofing.'''

    def __init__(self):
        self._root = None
        self._size = 0

    def add(self, e: Element):
        '''Add an element to the set. This is a no-op if the element is already
        in the set.

        :param e: the element to add'''
        if self._root is None:
            # we're the root, store here
            self._root = TreeNode(e)
            self._size = 1
        else:
            (added, r) = self._root.add(e)
            if added:
                self._size += 1
            if r is not None:
                # the tree was rotated about the root
                self._root = r

    def __contains__(self, e: Element) -> bool:
        '''Check whether the given element is a member of the locus.

        :param e: the element
        :returns: True if the element is in the locus'''
        if self._root is None:
            return False
        else:
            return self._root.find(e) is not None

    def empty(self) -> bool:
        '''Test if the locus is empty.

        :returns: True if the locus is empty'''
        return self._root is None

    def __len__(self) -> int:
        '''Return the size of the locus.

        :returns: the size of the locus'''
        return self._size

    def __iter__(self) -> Iterator[Element]:
        '''Return an iterator that returns elements in order.

        :returns: an iterator'''
        if self._root is None:
            return iter([])
        else:
            return iter(self._root)

    def discard(self, e: Element):
        '''Discard the given element from the locus. If the element
        isn't in the set, this is a no-op: use :meth:`remove`
        to detect removal of non-elements.

        :param e: the element'''
        if self._root is not None:
            (present, empty, r) = self._root.discard(e)
            if present:
                self._size -= 1
            if empty:
                # tree has been emptied
                self._root = None
            elif r is not None:
                # the tree was rotated about the root
                self._root = r

    def remove(self, e: Element):
        '''Remove the given element from the locus, raising
        an exception if it wasn't present: use :meth:`discard`
        to allow the removal of non-elements.

        :param e: the element'''
        if self._root is None:
            raise KeyError(e)
        else:
            (present, empty, r) = self._root.discard(e)
            if present:
                self._size -= 1
            else:
                raise KeyError(e)
            if empty:
                # tree has been emptied
                self._root = None
            elif r is not None:
                # the tree was rotated about the root
                self._root = r

    def draw(self) -> Element:
        '''Draw an element from the locus at random.

        :returns: a random element, or none if the set is empty'''
        if self._root is None:
            raise ValueError('Drawing from an empty locus')
        else:
            return self._root.draw()



# An AVL tree with random draw
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Bitstream, Element
from typing import List, Tuple, Iterator, Optional


class TreeNode():
    '''A node in an AVL tree. This is used within :class:`DrawSet`
    to maintain the contents of a set in a way that can perform a
    random fair draw of its contents.

    :param d: the data at the tree node
    :param p: (optional) the parent node (defaults to None)

    '''

    def __init__(self, d: Element, p: Optional['TreeNode'] = None):
        self._left: Optional['TreeNode'] = None    # left sub-tree
        self._right: Optional['TreeNode'] = None   # right sub-tree
        self._parent: Optional['TreeNode'] = p     # parent nodes
        self._data: Element = d                    # value at this node
        self._height: int = 0                      # height of the tree

    def __len__(self) -> int:
        '''Return the size of the tree. This is intended for testing only,
        as it's a slow recursive count.

        :returns: the size of the tree'''
        s = 1
        if self._left is not None:
            s += len(self._left)
        if self._right is not None:
            s += len(self._right)
        return s

    def add(self, e: Element) -> Tuple[bool, Optional['TreeNode']]:
        '''Add an element to the tree.

        This method performs two tasks. It adds the element, testing
        whether it was in fact added or was a repetition; and it
        potentially replaces trhe overall tree root.

        :returns: a pair (wasadded, newroot)'''
        if e == self._data:
            return (False, None)
        elif e < self._data:
            if self._left is None:
                self._left = TreeNode(e, self)
                self._updateHeights()
                return (True, self._left._rebalance())
            else:
                return self._left.add(e)
        else:
            if self._right is None:
                self._right = TreeNode(e, self)
                self._updateHeights()
                return (True, self._right._rebalance())
            else:
                return self._right.add(e)

    def _updateHeights(self):
        '''Walk back to the root updating the heights of nodes.'''
        if self._updateHeight() and self._parent is not None:
            self._parent._updateHeights()

    def _updateHeight(self) -> bool:
        '''Update the node height.

        :returns: True if the height of the node changed'''
        lh = self._left._height + 1 if self._left is not None else 0
        rh = self._right._height + 1 if self._right is not None else 0
        h = max(lh, rh)
        if h != self._height:
            self._height = h
            return True
        else:
            return False

    def isUnbalanced(self) -> bool:
        '''Test whether the node is unbalanced, defined as its
        sub-tree heights differing by more than one.

        :returns: True if the node is unbalanced'''
        lh = self._left._height + 1 if self._left is not None else 0
        rh = self._right._height + 1 if self._right is not None else 0
        return abs(lh - rh) > 1

    def _findUnbalanced(self) -> Optional['TreeNode']:
        '''Walk back up the tree looking for the shallowest unbalanced
        node.

        :returns: the shallowest unbalanced node, or None if the tree is completely balanced'''
        if self.isUnbalanced():
            return self
        elif self._parent is None:
            return None
        else:
            return self._parent._findUnbalanced()

    def _tallerSubtree(self) -> 'TreeNode':
        '''Return the taller of teh node's sub-trees.

        :returns: the smaller sub-tree'''
        lh = self._left._height + 1 if self._left is not None else 0
        rh = self._right._height + 1 if self._right is not None else 0
        if lh > rh:
            return self._left
        else:
            return self._right

    def _rebalance(self, recursive: bool = False) -> Optional['TreeNode']:
        '''Re-balance the tree after addition of a node. If the re-balancing
        is recursive (which is needed after deletions, but not after
        additions) then the re-balancing proceeds up the tree to the
        root.

        :param recursive: (optional) True to recursively re-balance (defaults to False)
        :returns: the new overall root, or None if the root hasn't been changed

        '''

        # find the shallowest unbalanced node
        z = self._findUnbalanced()
        if z is None:
            # tree is balanced all the way up
            return None
        else:
            root = z._rotate()

            if root._parent is None:
                return root
            elif recursive:
                # call again on the parent
                return root._parent._rebalance(True)
            else:
                return None

    def _rotate(z) -> Optional['TreeNode']:
        '''Perform a single rotation to re-balance the node.

        :returns: the new overall root, or None if the root hasn't been changed

        '''

        # we use z instead of the more usual self as the name for
        # the node on which the method is called, to match the
        # normal usage when describing AVL operations for which
        # z is always the root node.

        # find the two other nodes for the rotation
        y = z._tallerSubtree()
        x = y._tallerSubtree()

        # grab the parent to which we'll re-attach the new root
        # after rotation
        parent = z._parent

        # perform the appropriate rotation
        if z._data < y._data:
            if x._data < y._data:
                root = x
                #print('a-c-b about ' + str(z._data))
                y._left = x._right
                if x._right is not None:
                    x._right._parent = y
                z._right = x._left
                if x._left is not None:
                    x._left._parent = z
                x._left = z
                z._parent = x
                x._right = y
                y._parent = x
                z._updateHeight()
                y._updateHeight()
                if z.isUnbalanced():
                    z._rotate()
                if y.isUnbalanced():
                    y._rotate()
                x._updateHeight()
            else:
                root = y
                #print('a-b-c about ' + str(z._data))
                z._right = y._left
                if y._left is not None:
                    y._left._parent = z
                y._left = z
                z._parent = y
                z._updateHeight()
                if z.isUnbalanced():
                    z._rotate()
                y._updateHeight()
        elif x._data < y._data:
            root = y
            #print('c-b-a about ' + str(z._data))
            z._left = y._right
            if y._right is not None:
                y._right._parent = z
            y._left = x
            x._parent = y
            y._right = z
            z._parent = y
            z._updateHeight()
            if z.isUnbalanced():
                z._rotate()
            y._updateHeight()
        else:
            root = x
            #print('c-a-b about ' + str(z._data))
            y._right = x._left
            if x._left is not None:
                x._left._parent = y
            z._left = x._right
            if x._right is not None:
                x._right._parent = z
            x._left = y
            y._parent = x
            x._right = z
            z._parent = x
            z._updateHeight()
            y._updateHeight()
            if z.isUnbalanced():
                z._rotate()
            if y.isUnbalanced():
                y._rotate()
            x._updateHeight()

        # re-parent the new root
        root._parent = parent
        if parent is not None:
            # glue new local root into the parent in place of z,
            # the old root
            if parent._left == z:
                parent._left = root
            else:
                parent._right = root

            # update the heights back up to the root
            parent._updateHeights()

        # return the new root of the rotated tree
        return root

    def find(self, e: Element) -> 'TreeNode':
        '''Search for an element in the tree, returning its node..

        :param e: the element
        :returns: the node holding the element, or None'''
        if e == self._data:
            return self
        elif e < self._data:
            if self._left is None:
                return None
            else:
                return self._left.find(e)
        else:
            if self._right is None:
                return None
            else:
                return self._right.find(e)

    def __iter__(self) -> Iterator[Element]:
        '''Return an iterator that returns elements in order.

        :returns: an iterator'''
        return self._inOrder()

    def _inOrder(self) -> Element:
        '''Generator t oreturn the elements of the tree
        in order using an in-order traverse.

        :returns: the next element'''
        if self._left is not None:
            yield from self._left._inOrder()
        yield self._data
        if self._right is not None:
            yield from self._right._inOrder()

    def _leftmost(self) -> 'TreeNode':
        '''Return the leftmost node in a tree. This by definition
        holds the smallest element.

        :returns: the leftmost node'''
        if self._left is None:
            return self
        else:
            return self._left._leftmost()

    def _rightmost(self) -> 'TreeNode':
        '''Return the rightmost node in a tree. This by defintion holds
        the largest element.

        :returns: the rightmost node'''
        if self._right is None:
            return self
        else:
            return self._right._rightmost()

    def discard(self, e) -> Tuple[bool, bool, 'TreeNode']:
        '''Delete the given element from the tree, if it is present.

        This method performs three integrated tasks. If the element
        is present, it is removed; the resulting set is tested
        for emptiness; and any change to the overall root is detected.
        These three tasks each form an element of the return value.

        :returns: a triple (waspresent, nowempty, newroot)'''
        if e == self._data:
            # grab the parent node that will need to have
            # the replacement re-attached
            parent = self._parent

            # switch on the states of sub-trees
            if self._left is None:
                if self._right is None:
                    # leaf node, can be deleted immediately
                    #print('leaf')
                    if parent is None:
                        # we're the last node in the tree
                        return (True, True, None)
                    else:
                        # delete from parent
                        if parent._left == self:
                            #print('on left')
                            parent._left = None
                        else:
                            #print ('on right')
                            parent._right = None
                        parent._updateHeights()
                        return (True, False, parent._rebalance(True))
                else:
                    #print('right sub-tree only')
                    # only a right sub-tree, slide up to replace
                    if parent is None:
                        # we're the root, replace us
                        self._right._parent = None
                        return (True, False, self._right)
                    else:
                        # replace us with our sub-tree
                        if parent._left == self:
                            parent._left = self._right
                            self._right._parent = parent
                        else:
                            parent._right = self._right
                            self._right._parent = parent
                        parent._updateHeights()
                        return (True, False, parent._rebalance(True))
            elif self._right is None:
                #print('left sub-tree only')
                # only a left sub-tree, slide up to replace
                if self._parent is None:
                    # we're the root, replace us
                    self._left._parent = None
                    return (True, False, self._left)
                else:
                    # replace us with our sub-tree
                    if parent._left == self:
                        parent._left = self._left
                        self._left._parent = parent
                    else:
                        parent._right = self._left
                        self._left._parent = parent
                    parent._updateHeights()
                    return (True, False, parent._rebalance(True))
            else:
                #print('two sub-trees')
                # two sub-trees, choose the least disruptive element
                # from the taller as replacement
                if self._left._height > self._right._height:
                    r = self._left._rightmost()
                    #print('replace with ' + str(r._data))
                else:
                    r = self._right._leftmost()
                    #print('replace with ' + str(r._data))
                self._data = r._data
                return r.discard(r._data)

        elif e < self._data:
            if self._left is None:
                return (False, False, None)
            else:
                return self._left.discard(e)
        else:
            if self._right is None:
                return (False, False, None)
            else:
                return self._right.discard(e)

    def draw(self) -> Element:
        '''Draw an element from the tree at random.

        :returns: a random element, or none if the set is empty'''
        h = self._height
        if h == 0:
            return self._data

        bs = Bitstream.default_rng
        if self._left != None:
            if next(bs) == 0:
                return self._left.draw()

        if self._right is not None:
            for _ in range(h):
                if next(bs) == 1:
                    return self._right.draw()

        return self._data



# Test processes compose into sequences
#
# Copyright (C) 2021 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
from epyc import Experiment
import unittest
import networkx

class ProcessSequenceTest(unittest.TestCase):

    def testOneSequence(self):
        '''Test we can create a one-process sequence.'''
        params = dict()
        params[ERNetwork.N] = 1000
        params[ERNetwork.KMEAN] = 5
        params[SIR.P_INFECT] = 0.2
        params[SIR.P_REMOVE] = 0.2
        params[SIR.P_INFECTED] = 0.01

        # run as indiviaul process
        e = StochasticDynamics(SIR(), ERNetwork())
        rc1 = e.set(params).run()

        # run as single-process sequence
        p = ProcessSequence([SIR()])
        e = StochasticDynamics(p, ERNetwork())
        rc2 = e.set(params).run()

        self.assertAlmostEqual(rc1[Experiment.RESULTS][SIR.REMOVED], rc2[Experiment.RESULTS][SIR.REMOVED], delta=100)

    def testTwoSequence(self):
        '''Test we can cascade two processes.'''
        params = dict()
        params[ERNetwork.N] = 1000
        params[ERNetwork.KMEAN] = 5
        params[SIR.P_INFECT] = 0.2
        params[SIR.P_REMOVE] = 0.2
        params[SIR.P_INFECTED] = 0.01

        # run the two processes
        p = ProcessSequence([SIR(), NetworkStatistics()])
        e = StochasticDynamics(p, ERNetwork())
        rc1 = e.set(params).run()

        # check we get results from both
        self.assertIn(SIR.REMOVED, rc1[Experiment.RESULTS])
        self.assertIn(NetworkStatistics.LCC, rc1[Experiment.RESULTS])

    def testNetworkPassed(self):
        '''Test the second process sees the network as changed by the first.'''
        params = dict()
        params[ERNetwork.N] = 1000
        params[ERNetwork.KMEAN] = 5
        params[SIR.P_INFECT] = 0.2
        params[SIR.P_REMOVE] = 0.2
        params[SIR.P_INFECTED] = 0.01
        params[Percolate.T] = 0.5

        # run process singly to get baseline
        e = StochasticDynamics(NetworkStatistics(), ERNetwork())
        rc1 = e.set(params).run()

        # reduce the network first by percolation, then run again
        p = ProcessSequence([Percolate(), NetworkStatistics()])
        e = StochasticDynamics(p, ERNetwork())
        rc2 = e.set(params).run()

        # bond percolation shouold reduce the number of edges (only) 
        self.assertTrue(rc1[Experiment.RESULTS][NetworkStatistics.N] == rc2[Experiment.RESULTS][NetworkStatistics.N])
        self.assertTrue(rc1[Experiment.RESULTS][NetworkStatistics.M] > rc2[Experiment.RESULTS][NetworkStatistics.M])

if __name__ == '__main__':
    unittest.main()



# A source of random bits
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from typing import Optional, List, Iterable
import numpy


class Bitstream(object):
    '''An infinite stream of random bits. The bits are generated from a
    random number generated by the generator in `numpy` and consumed bitwise.

    :param size: (optional) size of the entropy pool in words'''

    # Singleton instance, created on demand
    default_rng: Optional['Bitstream'] = None    #: Default bit stream generator.

    @classmethod
    def init_default_rng(cl) -> 'Bitstream':
        '''Static method to return the default bitstream. This is called
        during system initialisation so that there is a generator always
        available.

        @returns: the bitstream'''
        if cl.default_rng is None:
            cl.default_rng = Bitstream()

    # Underlying types
    Dtype = numpy.int64                        #: Type for elements of the entropy pool.
    DtypeSize = 63                             #: Bits per element (excluding sign bit).

    def __init__(self, size: int = 100):
        self._rng = numpy.random.default_rng()

        self._pool: List[int] = []                      # entropy pool
        self._size: int = size                          # size of the pool
        self._max: int = 2 ** self.DtypeSize - 1        # maximum value of an entry in the pool
        self._element: int = 0                          # current element from the pool
        self._nelement: int = 0                         # index of current element
        self._index: int = 0                            # current bit within the element
        self._mask: int = 1                             # bit-mask

        self._refill()

    def _refill(self):
        '''Re-fill the entropy pool. This creates another batch of random numbers
        to be drawn from.'''
        self._pool = self._rng.integers(self._max, size=self._size, dtype=self.Dtype)
        self._nElement = 0
        self._element = int(self._pool[0])

    def __iter__(self) -> Iterable[int]:
        '''Return an iterator of bits.

        :retruns: an iterator'''
        return self

    def __next__(self) -> int:
        '''Return a random bit.

        :returns: a random bit'''
        bit = (self._element & self._mask) >> self._index
        self._index += 1
        self._mask <<= 1
        if self._index == self.DtypeSize:
            self._nElement += 1
            if self._nElement == self._size:
                self._refill()
            else:
                self._element = int(self._pool[self._nElement])
            self._index = 0
            self._mask = 1
        return bit



# Test process monitoring
#
# Copyright (C) 2020 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import unittest
import os
from tempfile import NamedTemporaryFile
import epyc
import networkx
from epydemic import *


class MonitoredSIR(SIR, Monitor):

    def __init__(self):
        super().__init__()


class MonitorTest(unittest.TestCase):

    def testSimple( self ):
        '''Test we capture the right time series.'''
        m = MonitoredSIR()
        m.setMaximumTime(100)
        e = StochasticDynamics(m, networkx.erdos_renyi_graph(1000, 5.0 / 1000))

        param = dict()
        param[SIR.P_INFECTED] = 0.01
        param[SIR.P_INFECT] = 0.002
        param[SIR.P_REMOVE] = 0.002
        param[Monitor.DELTA] = 1.0

        rc = e.set(param).run()
        self.assertSetEqual(set(rc[epyc.Experiment.RESULTS].keys()),
                            set([Monitor.timeSeriesForLocus(SIR.SI),
                                 Monitor.timeSeriesForLocus(SIR.INFECTED),
                                 Monitor.OBSERVATIONS,
                                 SIR.SUSCEPTIBLE,
                                 SIR.INFECTED,
                                 SIR.REMOVED]))
        # the next test is >=, not =, because some events may be drawn after the maxiumum time,
        # but the time is short enough that the number of infecteds won't be exhausted beforehand
        n = len(rc[epyc.Experiment.RESULTS][Monitor.OBSERVATIONS])
        self.assertGreaterEqual(n, 100)
        for k in [SIR.SI, SIR.INFECTED]:
            self.assertEqual(len(rc[epyc.Experiment.RESULTS][Monitor.timeSeriesForLocus(k)]), n)

    def testHDF5(self):
        '''Test we can save and retrieve the time series as HDF5.'''
        tf = NamedTemporaryFile()
        tf.close()
        fn = tf.name
        #fn = 'test.h5'

        try:
            nb = epyc.HDF5LabNotebook(fn, create=True)
            lab = epyc.Lab(nb)

            # run the experiment
            m = MonitoredSIR()
            m.setMaximumTime(100)
            e = StochasticDynamics(m, networkx.erdos_renyi_graph(1000, 5.0 / 1000))
            lab[SIR.P_INFECTED] = 0.01
            lab[SIR.P_INFECT] = 0.002
            lab[SIR.P_REMOVE] = 0.002
            lab[Monitor.DELTA] = 1.0
            rc = lab.runExperiment(e)
            df = lab.dataframe()

            # check we read back in correctly
            with epyc.HDF5LabNotebook(fn).open() as nb1:
                df1 = nb1.dataframe()
                r = df.iloc[0]
                r1 = df1.iloc[0]
                for f in [Monitor.OBSERVATIONS, Monitor.timeSeriesForLocus(SIR.SI), Monitor.timeSeriesForLocus(SIR.INFECTED)]:
                    self.assertCountEqual(r[f], r1[f])
        finally:
            try:
                os.remove(fn)
                #pass
            except OSError:
                pass


if __name__ == '__main__':
    unittest.main()



# Networks dynamics simulation base class
#
# Copyright (C) 2017--2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import sys
from heapq import heappush, heappop
from networkx import Graph
from epydemic import NetworkExperiment, Locus, Process, NetworkGenerator, EventFunction, EventDistribution, Element
if sys.version_info >= (3, 8):
    from typing import Union, Final, Dict, List, Any, Optional, Tuple, Callable
else:
    # backport compatibility with older typing
    from typing import Union, Dict, List, Any, Optional, Tuple, Callable
    from typing_extensions import Final

# Event types (not exported outside this file)
PostedEventFunction = Callable[[], None]
PostedEvent = Tuple[float, int, PostedEventFunction]


class Dynamics(NetworkExperiment):
    '''An abstract simulation framework for running a process over a network.

    This is the abstract base class for implementing different kinds
    of network process dynamics as computational experiments suitable
    for running under ``epyc``. Sub-classes provide synchronous and
    stochastic (Gillespie) simulation dynamics.

    The dynamics runs a network process provided as a :class:`Process`
    object. It is provided with a network generator that is called to
    generate a new experimental network instance for each run. The
    generator can be any iterator but will typically be an instance of
    :class:`NetworkGenerator`.

    :param p: network process to run
    :param g: (optional) prototype network or network generator'''

    # Additional metadata elements
    TIME: Final[str] = 'epydemic.Dynamics.time'      #: Metadata element holding the logical simulation end-time.
    EVENTS: Final[str] = 'epydemic.Dynamics.events'  #: Metadata element holding the number of events that happened.

    def __init__(self, p: Process, g: Union[Graph, NetworkGenerator] = None):
        super().__init__(g)

        # initialise other fields
        self._eventId: int = 0                                             # counter for posted events
        self._process: Process = p                                         # network process to run
        self._simulationTime: float = 0.0                                  # on-going simulation time
        self._process.setDynamics(self)                                    # back-link from process to dynamics (for events)
        self._loci: Dict[str, Locus] = dict()                              # dict from names to loci
        self._processLoci: Dict[Process, Dict[str, Locus]] = dict()        # dict from processes to loci for events
        self._perElementEvents: Dict[Process, EventDistribution] = dict()  # dict from processes to events that occur per-element
        self._perLocusEvents: Dict[Process, EventDistribution] = dict()    # dict from processes to events that occur per-locus
        self._postedEvents: List[PostedEvent] = []                         # pri-queue of fixed-time events


    # ---------- Configuration ----------

    def process(self) -> Process:
        '''Return the network process being run.

        :returns: the process'''
        return self._process

    def currentSimulationTime(self) -> float:
        '''Return the current simulation time.

        :returns: the current time'''
        return self._simulationTime

    def setCurrentSimulationTime(self, t: float):
        '''Set the current simulation time. This should only be used by
        sub-classes when running the simulation: doing so in any other
        context risks damaging the simulation.

        :param t: the new simualtion time

        '''
        self._simulationTime = t


    # ---------- Results ----------

    def experimentalResults(self) -> Dict[str, Any]:
        '''Report the process' experimental results. This simply calls through
        to the :meth:`Process.results` method of the process being
        simulated.

        :returns: the results of the process

        '''
        return self._process.results()


    # ---------- Set-up and tear-down ----------

    def setUp(self, params: Dict[str, Any]):
        '''Set up the experiment for a run. This performs the inherited
        actions, then builds the network process that the dynamics is
        to run.

        :params params: the experimental parameters

        '''
        super().setUp(params)

        # set up the event stream
        self._loci = dict()
        self._processLoci = dict()
        self._perElementEvents = dict()
        self._perLocusEvents = dict()
        self._postedEvents = []
        self._eventId = 0
        self._simulationTime = 0.0

        # build and set up the process
        self._process.reset()
        self._process.build(params)
        self._process.setUp(params)

    def tearDown(self):
        '''At the end of each experiment, throw away any posted by un-executed
        events.

        '''
        self._process.tearDown()
        super().tearDown()
        self._postedEvents= []


    # ---------- Stochastic events (drawn from a distribution) ----------

    def addLocus(self, p: Process, n: str, l: Locus = None):
        """Add a named locus associated with the given process.

        :param p: the process
        :param n: the locus name
        :param l: the locus (defaults to a simple set-based locus)
        :returns: the locus"""
        if n in self._loci.keys():
            raise Exception("Locus {n} already exists in the simulation".format(n = n))

        # store locus by name
        if l is None:
            l = Locus(n)
        self._loci[n] = l

        # update process record
        if p not in self._processLoci:
            # new process, add loci and event lists
            self._processLoci[p] = dict()
            self._perElementEvents[p] = []
            self._perLocusEvents[p] = []
        self._processLoci[p][n] = l

        # return the locus
        return l

    def locus(self, n: str) -> Locus:
        '''Retrieve a locus by name.

        :param n: the locus name
        :returns: the locus'''
        return self._loci[n]

    def loci(self) -> Dict[str, Locus]:
        '''Return all the loci in the simulation.

        :returns: a dict from names to loci'''
        return self._loci

    def lociForProcess(self, p: Process) -> Dict[str, Locus]:
        '''Return all the loci defined for a specific process.

        :param p: the process
        :returns: a dict from names to loci'''
        if p in self._processLoci:
            # process has loci, return them
            return self._processLoci[p]
        else:
            # process doesn't have loci, return an empty dict
            return dict()

    def addEventPerElement(self, p: Process, l: Union[str, Locus], pr: float, ef: EventFunction):
        """Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability for each element of the locus, and calling the :term:`event function`
        when it is selected.

        :param p: the process
        :param l: the locus or locus name
        :param pr: the event probability
        :param ef: the event function"""
        if isinstance(l, str):
            l = self.locus(l)
        self._perElementEvents[p].append((l, pr, ef))

    def perElementEventDistribution(self, p: Process) -> EventDistribution:
        """Return the distribution of per-element events for the given process' loci.

        :param p: the process
        :returns: a list of (locus, probability, event function) triples"""
        if p in self._perElementEvents:
            return self._perElementEvents[p]
        else:
            return []

    def addFixedRateEvent(self, p: Process, l: Union[str, Locus], pr: float, ef: EventFunction):
        """Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability, and calling the :term:`event function` when it is selected.

        :param p: the process
        :param l: the locus or locus name
        :param pr: the event probability
        :param ef: the event function"""
        if isinstance(l, str):
            l = self.locus(l)
        self._perLocusEvents[p].append((l, pr, ef))

    def fixedRateEventDistribution(self, p: Process) -> EventDistribution:
        """Return the distribution of fixed-rate events for the given process' loci.

        :param p: the process
        :returns: a list of (locus, probability, event function) triples"""
        if p in self._perElementEvents:
            return self._perLocusEvents[p]
        else:
            return []

    def eventRateDistribution(self, t: float) -> EventDistribution:
        """Return the event distribution, a sequence of (l, r, f) triples
        where l is the locus where the event occurs, r is the rate at
        which an event occurs, and f is the event function called to
        make it happen.

        Note the distinction between a *rate* and a *probability*:
        the former can be obtained from the latter simply by
        multiplying the event probability by the number of times it's
        possible in the current network, which for per-element events
        is the population of nodes or edges in a given state.

        It is perfectly fine for an event to have a zero rate. The process
        is assumed to have reached equilibrium if all events have zero rates.

        :param t: current time
        :returns: a list of (locus, rate, event function) triples"""
        rates = []

        for p in self._perElementEvents:
            # convert per-element events to rates
            for (l, pr, ef) in self._perElementEvents[p]:
                rates.append((l, pr * len(l), ef))

            # add fixed-rate events for non-empty loci
            for (l, pr, ef) in self._perLocusEvents[p]:
                if len(l) > 0:
                    rates.append((l, pr, ef))

        return rates


    # ---------- Posted events (occurring at a fixed time) ----------

    def _nextEventId(self) -> int:
        """Generate a sequence number for a posted event. This is used to ensure that
        all event triples pushed onto the priqueue are unique in their first two elements,
        and therefore never try to do comparisons with functions (the third element).

        :returns: a sequence number"""
        id = self._eventId
        self._eventId += 1
        return id

    def postEvent(self, t: float, e: Element, ef: EventFunction):
        """Post an event that calls the :term:`event function` at time t.

        :param t: the current time
        :param e: the element (node or edge) on which the event occurs
        :param ef: the event function"""
        heappush(self._postedEvents, (t, self._nextEventId(), (lambda: ef(t, e))))

    def postRepeatingEvent(self, t: float, dt: float, e: Element, ef: EventFunction):
        """Post an event that starts at time t and re-occurs at interval dt.

        :param t: the start time
        :param dt: the interval
        :param e: the element (node or edge) on which the event occurs
        :param ef: the element function"""

        def repeat(tc, e):
            ef(tc, e)
            tp = tc + dt
            heappush(self._postedEvents, (tp, self._nextEventId(), (lambda: repeat(tp, e))))

        heappush(self._postedEvents, (t, self._nextEventId(), (lambda: repeat(t, e))))

    def nextPendingEventBefore(self, t: float) -> Optional[PostedEventFunction]:
        """Return the next pending event to occur at or before time t.

        :param t: the current time
        :returns: a pending event function or None"""
        if len(self._postedEvents) > 0:
            # we have events, grab the soonest
            (et, id, pef) = heappop(self._postedEvents)
            if et <= t:
                # event should have occurred, return
                return pef
            else:
                # this (and therefore all further events) are in the future, put it back
                heappush(self._postedEvents, (et, id, pef))
                return None
        else:
            # we don't have any events
            return None

    def runPendingEvents(self, t: float) -> int:
        '''Retrieve and fire any pending events at time t. This handles
        the case where firing an event posts another event that needs to be run
        before other already-posted events coming before time t: in other words,
        it ensures that the simulation order is respected.

        :param t: the current time
        :returns: the number of events fired'''
        n = 0
        while True:
            pef = self.nextPendingEventBefore(t)
            if pef is None:
                # no more pending events, return however many we've fired already
                return n
            else:
                # fire the event
                pef()
                n += 1



# Helper types
#
# Copyright (C) 2017--2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from typing import Union, Tuple, Any

# the parts of a network (graph)
Node = Any                   #: A node in a network, which may be any object.
Edge = Tuple[Node, Node]     #: An edge in a network, connecting exactly two nodes (which may be the same).
Element = Union[Node, Edge]  #: An element in a simulation, either a node or an edge.





# Synchronous dynamics base class
#
# Copyright (C) 2017--2020 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import sys
from copy import copy
import numpy                     # type: ignore
from networkx import Graph
from epydemic import Dynamics, Process, NetworkGenerator
if sys.version_info >= (3, 8):
    from typing import Any, Dict, Union, Final
else:
    # backport compatibility with older typing
    from typing import Any, Dict, Union
    from typing_extensions import Final


class SynchronousDynamics(Dynamics):
    '''A dynamics that runs synchronously in discrete time, applying local
    rules to each node in the network. These are simple to understand and
    simple to code for many cases, but can be statistically inexact and slow
    for large systems.

    :pafram p: the network process to run
    :param g: prototype network to run over (optional)'''

    # additional metadata
    TIMESTEPS_WITH_EVENTS : Final[str] = 'epydemic.dynamics.timesteps_with_events'  #: Metadata element holding the number timesteps that actually had events occur within them

    def __init__(self, p : Process, g : Union[Graph, NetworkGenerator] =None):
        super().__init__(p, g)

    def do(self, params : Dict[str, Any]) -> Dict[str, Any]:
        '''Execute the process under synchronous dynamics.

        :param params: the parameters of the simulation
        :returns: a dict of experimental results'''

        # run the
        rng = numpy.random.default_rng()
        proc = self.process()
        t = 1.0
        events = 0
        timestepEvents = 0
        while not proc.atEquilibrium(t):
            self.setCurrentSimulationTime(t)

            # fire any events posted for at or before this time
            nev = self.runPendingEvents(t)

            # run through all the events in the distribution
            dist = self.perElementEventDistribution(t)
            for (l, p, ef) in dist:
                if (len(l) > 0) and (p > 0.0):
                    # run through every possible element on which this event may occur
                    for e in copy(l.elements()):
                        # test for occurrance of the event on this element
                        if rng.random() <= p:
                            # yes, perform the event
                            ef(t, e)
                            nev = nev + 1

            # run through all the fixed-rate events for this timestep
            dist = self.fixedRateEventDistribution(t)
            for (l, p, ef) in dist:
                if (len(l) > 0) and (p > 0.0):
                    if rng.random() <= p:
                        # yes, perform the event on an element drawn at random
                        e = l.draw()
                        ef(t, e)
                        nev = nev + 1

            # add the events to the count
            events = events + nev
            if nev > 0:
                # we had events happen in this timestep
                timestepEvents = timestepEvents + 1

            # advance to the next timestep
            t = t + 1.0

        # add some more metadata
        (self.metadata())[self.TIME] = t
        (self.metadata())[self.EVENTS] = events
        (self.metadata())[self.TIMESTEPS_WITH_EVENTS] = timestepEvents

        # report results
        rc = self.experimentalResults()
        return rc



# Process combinator to build sequences of processes
#
# Copyright (C) 2017--2021 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import Process, Dynamics
from typing import List, Dict, Any

class ProcessSequence(Process):
    '''A process build from a sequence of other processes. This allows separate process
    behaviour to be defined independently and then combined in different ways.
    
    :param ps: the processes'''

    def __init__(self, ps : List[Process]):
        self._processes = ps
        super(ProcessSequence, self).__init__()
        
    def setDynamics(self, d : Dynamics):
        '''Set the dynamics.

        :param d: the dynamics'''
        super(ProcessSequence, self).setDynamics(d)
        for p in self._processes:
            p.setDynamics(d)

    def reset(self):
        '''Reset the processes.'''
        for p in self._processes:
            p.reset()
            
    def build(self, params : Dict[str, Any]):
        '''Build the proceses.
        
        :param params: the experimental parameters'''
        for p in self._processes:
            p.build(params)
            
    def setUp(self, params : Dict[str, Any]):
        '''Set up the proceses.
        
        :param params: the experimental parameters'''
        for p in self._processes:
            p.setUp(params)
            
    def tearDown(self):
        '''Tear down the processes.'''
        for p in self._processes:
            p.tearDown()
            
    def atEquilibrium(self, t : float):
        '''Test for equilibrium. A process sequence is at equilibrium if and when all
        its component processes are.
        
        :param t: the simulation time
        :returns: True if all the processes are at equilibrium'''
        for p in self._processes:
            if not p.atEquilibrium(t):
                return False
        return True

    def setMaximumTime(self, t : float):
        '''Set the maximum default simulation time for all processes

        :param t: the maximum simulation time'''
        for p in self._processes:
            p.setMaximumTime(t)

    def maximumTime(self) -> float:
        '''Return the maximum assumed simulation time, which is the maximum
        of all component processes.

        :returns: the maximum simulation time'''
        t = 0
        for p in self._processes:
            t = max(t, p.maximumTime())
        return t

    def results(self) -> Dict[str, Any]:
        '''Return all the experimental results from all the processes. The dict
        is created in process order, meaning that later processes may alter or overwrite
        the results of earlier ones, accidentally or deliberately.
        
        :returns: a dict of experimental results'''
        res = super(ProcessSequence, self).results()
        for p in self._processes:
            res.update(p.results())
        return res
        


# Test balanced binary trees and their DrawSet wrapper
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import statistics
import numpy
import unittest


class BBTTest(unittest.TestCase):

    def assertInvariant(self, s):
        '''Test the invariant of tree is satisfied.'''
        if s._root is not None:
            self.assertAVLInvariant(s._root)

    def assertAVLInvariant(self, n):
        lh = n._left._height + 1 if n._left is not None else 0
        rh = n._right._height + 1 if n._right is not None else 0
        self.assertTrue(abs(lh - rh) <= 1)
        self.assertEqual(n._height, max(lh, rh))
        if lh > 0:
            self.assertIsNotNone(n._left)
            self.assertAVLInvariant(n._left)
        if rh > 0:
            self.assertIsNotNone(n._right)
            self.assertAVLInvariant(n._right)

    def testEmptySet(self):
        '''Test empty sets.'''
        s = DrawSet()
        self.assertInvariant(s)
        self.assertEqual(len(s), 0)
        self.assertFalse(1 in s)
        self.assertCountEqual(s, [])
        self.assertTrue(s.empty())

    def testAdd(self):
        '''Test we can add individual elements.'''
        s = DrawSet()
        self.assertTrue(s.empty())
        s.add(1)
        self.assertFalse(s.empty())
        self.assertInvariant(s)
        self.assertEqual(len(s), 1)
        self.assertTrue(1 in s)
        s.add(2)
        self.assertInvariant(s)
        self.assertEqual(len(s), 2)
        self.assertTrue(1 in s)
        self.assertTrue(2 in s)
        self.assertCountEqual(list(s), [1,2])
        s.add(3)
        self.assertInvariant(s)
        self.assertEqual(len(s), 3)
        self.assertTrue(1 in s)
        self.assertTrue(2 in s)
        self.assertTrue(3 in s)
        self.assertCountEqual(list(s), [1, 2, 3])

    def testABCroot(self):
        '''Test the ABC rotation over the root.'''
        s = DrawSet()
        s.add(0)
        s.add(1)
        s.add(2)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [0, 1, 2])
        self.assertEqual(s._root._data, 1)

    def testABC(self):
        '''Test the ABC rotation below the root.'''
        s = DrawSet()
        s.add(1)
        s.add(0)
        s.add(2)
        s.add(3)
        s.add(4)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [0, 1, 2, 3, 4])
        self.assertEqual(s._root._data, 1)

    def testACBroot(self):
        '''Test the ACB rotation overt the root.'''
        s = DrawSet()
        s.add(1)
        s.add(3)
        s.add(2)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [1, 2, 3])
        self.assertEqual(s._root._data, 2)

    def testACB(self):
        '''Test the ACB rotation below the root.'''
        s = DrawSet()
        s.add(3)
        s.add(0)
        s.add(5)
        s.add(7)
        s.add(6)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [0, 3, 5, 6, 7])
        self.assertEqual(s._root._data, 3)

    def testCBAroot(self):
        '''Test the CBA rotation across the root.'''
        s = DrawSet()
        s.add(3)
        s.add(2)
        s.add(1)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [1, 2, 3])
        self.assertEqual(s._root._data, 2)

    def testCBA(self):
        '''Test the CBA rotation below the root.'''
        s = DrawSet()
        s.add(5)
        s.add(7)
        s.add(3)
        s.add(2)
        s.add(1)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [1, 2, 3, 5, 7])
        self.assertEqual(s._root._data, 5)

    def testCABroot(self):
        '''Test the CABB rotation across the root.'''
        s = DrawSet()
        s.add(4)
        s.add(2)
        s.add(3)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [2, 3, 4])
        self.assertEqual(s._root._data, 3)

    def testCAB(self):
        '''Test the CAB rotation below the root.'''
        s = DrawSet()
        s.add(2)
        s.add(1)
        s.add(8)
        s.add(5)
        s.add(6)
        self.assertInvariant(s)
        self.assertCountEqual(list(s), [1, 2, 5, 6, 8])
        self.assertEqual(s._root._data, 2)

    def testDiscard(self):
        '''Test we can discard elements that are present.'''
        s = DrawSet()
        s.add(2)
        s.add(3)
        s.discard(3)
        self.assertInvariant(s)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])

    def testDiscardEmpty(self):
        '''Test we can discard from an empty set.'''
        s = DrawSet()
        s.discard(1)
        self.assertInvariant(s)
        self.assertEqual(len(s), 0)
        self.assertCountEqual(s, [])

    def testRemoveSuccess(self):
        '''Test successful removal.'''
        s = DrawSet()
        s.add(1)
        s.add(2)
        s.remove(2)
        self.assertInvariant(s)

    def testRemoveFailure(self):
        '''Test failing removal of a non-element.'''
        s = DrawSet()
        s.add(1)
        s.add(2)
        with self.assertRaises(KeyError):
            s.remove(3)

    def testRemoveEmpty(self):
        '''Test removal; from an empty set.'''
        s = DrawSet()
        with self.assertRaises(KeyError):
            s.remove(3)

    def testDiscardToEmpty(self):
        '''Test we can discard the last element in the set.'''
        s = DrawSet()
        s.add(1)
        s.discard(1)
        self.assertInvariant(s)
        self.assertEqual(len(s), 0)
        self.assertCountEqual(s, [])
        self.assertTrue(s.empty())

    def testRemoveToEmpty(self):
        '''Test we can removethe last element in the set.'''
        s = DrawSet()
        s.add(1)
        s.remove(1)
        self.assertInvariant(s)
        self.assertEqual(len(s), 0)
        self.assertCountEqual(s, [])
        self.assertTrue(s.empty())

    def testDiscardLeaves(self):
        '''Test we can discard leaf elements.'''
        s = DrawSet()
        s.add(2)
        s.add(3)
        s.discard(3)
        self.assertInvariant(s)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])
        s.add(1)
        s.discard(1)
        self.assertInvariant(s)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])
        s.discard(2)
        self.assertInvariant(s)
        self.assertEqual(len(s), 0)
        self.assertCountEqual(s, [])

    def testDiscardSingleSubtree(self):
        '''Test we can discard nodes with a single sub-tree.'''
        s = DrawSet()
        s.add(2)
        s.add(3)
        s.add(4)
        s.add(5)
        s.discard(4)
        self.assertInvariant(s)
        self.assertEqual(len(s), 3)
        self.assertCountEqual(s, [2, 3, 5])
        s.add(1)
        s.discard(2)
        self.assertInvariant(s)
        self.assertEqual(len(s), 3)
        self.assertCountEqual(s, [1, 3, 5])

    def testDiscardTwoSubtrees(self):
        '''Test we can discard a node with two sub-trees.'''
        s = DrawSet()
        s.add(3)
        s.add(2)
        s.add(10)
        s.add(7)
        s.add(15)
        s.add(6)
        s.discard(10)
        self.assertInvariant(s)
        self.assertEqual(len(s), 5)
        self.assertCountEqual(s, [2, 3, 6, 7, 15])

    def testRepeatedAdd(self):
        '''Test we can add the same element twice.'''
        s = DrawSet()
        s.add(1)
        self.assertInvariant(s)
        s.add(2)
        self.assertInvariant(s)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])
        s.add(2)
        self.assertInvariant(s)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])
        s.add(1)
        self.assertInvariant(s)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])

    def testNonDiscard(self):
        '''Test we can discard when the element isn't there.'''
        s = DrawSet()
        s.add(1)
        s.add(2)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])
        s.discard(3)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])
        s.discard(1)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])
        s.discard(2)
        self.assertEqual(len(s), 0)
        self.assertCountEqual(s, [])

    def testRepeatedDiscard(self):
        '''Test repeated discarding.'''
        s = DrawSet()
        s.add(1)
        s.add(2)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])
        s.discard(1)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])
        s.discard(1)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])

    def testRepeatedRemove(self):
        '''Test repeated removing.'''
        s = DrawSet()
        s.add(1)
        s.add(2)
        self.assertEqual(len(s), 2)
        self.assertCountEqual(s, [1, 2])
        s.remove(1)
        self.assertEqual(len(s), 1)
        self.assertCountEqual(s, [2])
        with self.assertRaises(KeyError):
            s.remove(1)

    def testLotsOfAdds(self):
        '''Test we can add loads of elements, and get the results in order.'''
        s = DrawSet()
        es = list(range(200))
        rs = es.copy()
        numpy.random.shuffle(es)
        for i in es:
            s.add(i)
            self.assertInvariant(s)
        self.assertEqual(len(s), len(rs))
        self.assertCountEqual(rs, list(s))

    def testAddRemoveInOrder(self):
        '''Test we can handle an maximally inbalanced tree.'''
        s = DrawSet()
        es = list(range(20))
        for i in es:
            s.add(i)
            self.assertInvariant(s)
        self.assertEqual(len(s), len(es))
        for i in es:
            s.remove(i)
            self.assertInvariant(s)
        self.assertEqual(len(s), 0)

    def testAddRemoveInReverseOrder(self):
        '''Test we can handle an maximally inbalanced tree the other way.'''
        s = DrawSet()
        es = list(range(100))
        es.reverse()
        for i in es:
            s.add(i)
            self.assertInvariant(s)
        self.assertEqual(len(s), len(es))
        for i in es:
            s.remove(i)
            self.assertInvariant(s)
        self.assertEqual(len(s), 0)

    def testAddRemove(self):
        '''Test we can add in random order and then remove in order.'''
        s = DrawSet()
        es = list(range(100))
        rs = es.copy()
        numpy.random.shuffle(rs)
        for i in rs:
            s.add(i)
            self.assertInvariant(s)
        l = len(s)
        for j in es:
            s.remove(j)
            self.assertInvariant(s)
            l -= 1
            self.assertEqual(len(s), l)
        self.assertEqual(len(s), 0)

    def testDraw(self):
        '''Test the basic draw functionality.'''
        s = DrawSet()
        es = list(range(1000))
        numpy.random.shuffle(es)
        for e in es:
            s.add(e)
        for _ in range(int(len(es) * 0.1)):
            j = s.draw()
            self.assertIn(j, s)

    def testDrawAll(self):
        '''Test we always draw all elements.'''
        s = DrawSet()
        es = list(range(1000))
        numpy.random.shuffle(es)
        for e in es:
            s.add(e)
        ns = []
        for _ in range(len(es)):
            n = s.draw()
            self.assertIsNotNone(n)
            self.assertNotIn(n, ns)
            ns.append(n)
            s.remove(n)
            self.assertInvariant(s)
            self.assertNotIn(n, s)
        self.assertEqual(len(s), 0)
        self.assertCountEqual(es, ns)


if __name__ == '__main__':
    unittest.main()



# Monitor the progress of an epidemic
#
# Copyright (C) 2020--2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import sys
from typing import Dict, Any
if sys.version_info >= (3, 8):
    from typing import Final
else:
    # backport compatibility with older typing
    from typing_extensions import Final
from epydemic import Process


class Monitor(Process):
    '''Add progress monitoring to a process. This class captures the sizes
    of all loci in the process at regular intervals, returning them as
    time series. Each time series appears as an array in the results dict, keyed
    by :attr:`OBSERVATIONS` (for the simulation times at which the observations
    were taken) or by a name related to the locus name, generated by
    :meth:`timeSeriesForLocus`. This means that each locus has its own
    time series tracking its evolving size.

    '''

    # Experimental parameters
    DELTA: Final[str] = "epydemic.monitor.time_delta"             #: Parameter for the time interval for observations.

    # Results
    OBSERVATIONS: Final[str] = 'epydemic.monitor.observations'    #: Result holding the times of the observations.

    # Result name stem
    TIMESERIES_STEM: Final[str] = "epydemic.monitor.timeseries"   #: Stem for names of the timeseries for different loci.

    @staticmethod
    def timeSeriesForLocus(l: str) -> str:
        '''Generate the name of the time series corresponding to the
        given locus.

        :param l: the locus
        :returns: the result name'''
        return '{stem}-{locus}'.format(stem=Monitor.TIMESERIES_STEM, locus=l)

    def __init__(self):
        super().__init__()

    def reset(self):
        '''Reset the process.'''
        super().reset()
        self._timeSeries = None

    def build(self, params: Dict[str, Any]):
        '''Build the observation process.

        :param params: the experimental parameters

        '''
        super().build(params)

        # post a repeating event to observe the process
        delta = params[self.DELTA]
        self.postRepeatingEvent(0.0, delta, None, self.observe)

    def observe(self, t: float, e: Any):
        '''Observe the network, capturing the sizes of all loci which are then
        stored into individual time series. Another time series,
        tagged :attr:`OBSERVATIONS`, is created to store the sequence
        of times at which observations were made.

        :param t: the current simulation time
        :param e: the element (ignored)

        '''

        # if this is the first run, build the initial dicts: one
        # per locus and one for the observation times
        # (We can't do this as part of build() as it depends on which
        # other processes we're composed with, and in what order)
        if self._timeSeries is None:
            self._timeSeries = dict()
            self._timeSeries[self.OBSERVATIONS] = []
            for n in self.dynamics().loci().keys():   # loci for all processes in this experiment
                self._timeSeries[Monitor.timeSeriesForLocus(n)] = []

        # make the observation
        self._timeSeries[self.OBSERVATIONS].append(t)
        for (n, l) in self.dynamics().loci().items():
            self._timeSeries[Monitor.timeSeriesForLocus(n)].append(len(l))

    def results(self) -> Dict[str, Any]:
        '''Return the time series tagged as specified by
        :meth:`timeSeriesForLocus`. There is one time series *per*
        locus, plus one for the sequence of times at which the
        observations were made.

        :returns: the results

        '''
        res = super().results()

        # store the time series
        for n in self._timeSeries:
            res[n] = self._timeSeries[n]
        return res



# Test basic loci functions
#
# Copyright (C) 2017--2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import unittest


class LociTest(unittest.TestCase):

    def testName(self):
        '''Test name retrieval.'''
        l = Locus('test')
        self.assertEqual(l.name(), 'test')

    def testAdd(self):
        '''Test we can add elements.'''
        l = Locus('test')
        self.assertEqual(len(l), 0)
        l.add(1)
        self.assertEqual(len(l), 1)
        l.add(2)
        self.assertEqual(len(l), 2)

    def testDelete(self):
        '''Test we can remove elements.'''
        l = Locus('test')
        l.add(1)
        l.add(2)
        l.discard(1)
        self.assertEqual(len(l), 1)
        l.discard(2)
        self.assertEqual(len(l), 0)

    def testDrawEmpty(self):
        '''Test we can't draw from an empty locus.'''
        l = Locus('test')
        with self.assertRaises(ValueError):
            l.draw()

    def testDraw(self):
        '''Test the basic draw functionality.'''
        l = Locus('test')
        es = list(range(100))
        for i in es:
            l.add(i)
        j = l.draw()
        self.assertIn(j, l)

    def testDrawOne(self):
        '''Test we can draw an element without changing the locus.'''
        l = Locus('test')
        l.addHandler(None, 1)
        l.addHandler(None, 2)
        self.assertEqual(len(l), 2)
        self.assertCountEqual(l, [1, 2])

        # duplicate entries are ignored
        l.addHandler(None, 2)
        self.assertCountEqual(l, [1, 2])

        # drawing leaves the population unchanged
        e = l.draw()
        self.assertIn(e, [1, 2])
        self.assertCountEqual(l, [1, 2])


if __name__ == '__main__':
    unittest.main()



# The Newman-Ziff algorithm for site (node) and bond (edge) percolation
#
# Copyright (C) 2017--2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import NetworkExperiment, Node, Edge
from epyc import Experiment, ResultsDict
from networkx import Graph
import numpy
import sys
from typing import Any, Dict, Union, Optional, Iterable, List, cast
if sys.version_info >= (3, 8):
    from typing import Final
else:
    # backport compatibility with older typing
    from typing_extensions import Final


class NewmanZiff(NetworkExperiment):
    '''Base class for the Newman-Ziff site and bond percolation algorithm,
    as described in :ref:`Newman and Ziff 2000 <NZ00>`.
    
    :param g: (optional) the underlying network or generator
    :param samples: (optional) number of samples or list of sample points (defaults to 100)'''

    def __init__(self, g : Graph =None, samples : Union[int, Iterable[float]] =None):
        super(NewmanZiff, self).__init__(g)

        # fill in default
        if samples is None:
            samples = 100
        if isinstance(samples, int):
            samples = numpy.linspace(0.0, 1.0, num=samples, endpoint=True)
        self._samples : List[float] = sorted(numpy.unique(list(cast(Iterable[float], samples))))

        # components data structure is initially empty
        self._components : numpy.ndarray = None

    def tearDown(self):
        '''Throw away the components data structure at tear-down.'''
        self._components = None
        super(NewmanZiff, self).tearDown()

    def rootOf(self, n : Node) -> Node:
        '''Return the root of the component containing node n, updating
        the tree accordingly.

        :param n: the node in the component
        :returns: the root of the component'''
        np = self._components[n]
        if np < 0:
            # n is the root, return it
            return n
        else:
            # n has a parent, follow the tree to it
            r = self.rootOf(np)

            # update our component record to point to the root
            self._components[n] = r

            # return the root
            return r

    def join(self, c1 : Node, c2 : Node) -> int:
        '''Join two components (as represented by their roots), returning
        the size of the combined component. The root of the combined
        component is c1.

        :param c1: the first component root
        :param c2: trhe second component root
        :returns: the new component's size'''
        # extract the size of the second compooent
        msize = self._components[c2]
        
        # join the second compoent to the first 
        self._components[c2] = c1
        
        # update the size of the first component
        self._components[c1] += msize
                
        # return the size of the new component
        return -self._components[c1]        

    def sample(self, p : float) -> Dict[str, Any]:
        '''Take a sample. The default does nothing.

        :param p: the current occupation probability
        :returns: an empty dict'''
        return dict()

    def report(self, params : Dict[str, Any], meta : Dict[str, Any], res : List[Dict[str, Any]]) -> List[ResultsDict]:
        '''Re-write the list of results into a list of individual experiments by
        wrapping them up as results dicts. Essentially this presents the results as though
        a percolation experiment had happened for each sample.

        :param params: the experimental parameters
        :param meta: the metadata
        :param res: a list of individual sample results
        :returns: a list of results dicts'''
        rcs = []
        for r in res:
            # wrap the experiment results in their own results dict
            rc = self.resultsdict()
            rc[self.PARAMETERS] = params.copy()
            rc[self.METADATA] = meta.copy()
            rc[self.RESULTS] = r.copy()
            rcs.append(rc)
        return rcs


class BondPercolation(NewmanZiff):
    '''A bond (edge) percolation experiment. This experiment computes the size of
    the giant connected component (GCC) as edges are "occupied" within the underlying network.
    It samples the GCC at a given sequence of occupation probabilities, returning
    a time series of the growth of the GCC.

    :param g: (optional) the underlying network or generator
    :param samples: (optional) number of samples or list of sample points (defaults to 100)'''

    # Synthesised parameters
    P : Final[str] = 'epydemic.bondpercolation.pOccupied'    #: Parameter holding percolation threshold.

    # Experimental results
    GCC : Final[str] = 'epydemic.bondpercolation.gcc'        #: Result holding size of GCC.

    def __init__(self, g : Graph =None, samples : Union[int, Iterable[float]] =None):
        super(BondPercolation, self).__init__(g, samples)

    def setUp(self, params : Dict[str, Any]):
        '''Set up the process, creating the initial components data structure from the
        underlying network.

        :param params: the experimental parameters'''
        super(BondPercolation, self).setUp(params)
        N = self.network().order()
        self._components = numpy.full(N, -1, numpy.int32)
        self._gcc = 1   # initially all nodes are individual components, unconnected by occupied edges

    def occupy(self, n : Node, m : Node) -> Optional[int]:
        '''Occupy an edge. If this causes two components to be joined, 
        update the GCC and return the size of the new component; otherwise 
        return None.

        This method should be overridden to collect more statistics about the network
        as components join.

        :param n: one node
        :param m: the other nodse
        :returns: the size of any newly-combined component, or None'''
        nr = self.rootOf(n)
        mr = self.rootOf(m)
        if mr != nr:
            # nodes are in different components, join them together
            csize = self.join(nr, mr)

            # update the GCC
            self._gcc = max(self._gcc, csize)

            # return the new component size
            return csize
        else:
            # nodes are in the same component, do nothing
            return None

    def sample(self, p : float) -> Dict[str, Any]:
        '''Take a sample. The default samples the size of the GCC.

        :param p: the current occupation probability
        :returns: a dict of results'''
        res = super(BondPercolation, self).sample(p)
        res[self.P] = p        
        res[self.GCC] = self._gcc        
        return res

    def percolate(self, es : List[Edge]) -> List[Dict[str, Any]]:
        '''Perform the bond percolation process. This runs through the
        list of edges, occupying them and taking samples of the occupied
        network's statistics at the requested sample points. The samples
        will be taken as the closest point above the requested probability.
        They will be recorded in the results as actually having been sampled at the
        requested point, however, so that all results are sampled with the same indices.

        :param es: the permuted list of edges
        :returns: a list of dicts of experiment results.'''
        # take an initial sample if requested
        samples = []
        samplePoint = 0
        if self._samples[samplePoint] == 0.0:
            samples.append(self.sample(self._samples[samplePoint]))
            samplePoint += 1

        # percolate the network
        M = len(es)
        for i in range(M):
            (n, m) = es[i]

            # occupy the edge
            self.occupy(n, m)

            # take a sample if this is a sample point
            if  (i + 1) / M >= self._samples[samplePoint]:
                # we're at the closest probability after the requested sample point,
                # so build the sample
                samples.append(self.sample(self._samples[samplePoint]))

                # if we've collected all the samples we want, bail out
                samplePoint += 1
                if samplePoint > len(self._samples):
                    break
                    
        return samples

    def do(self, params : Dict[str, Any]) -> List[Dict[str, Any]]:
        '''Perform the percolation process. This passes a permuted
        list of edges to :meth:`NewmanZiff.percolate`, which performs
        the actual operation.

        :param params: experimental parameters
        :returns: a list of dicts of experimental results'''        
        # extract and shuffle the edges
        g = self.network()
        es = list(g.edges()).copy()
        numpy.random.shuffle(es)

        # percolate the network using these edges
        return self.percolate(es)


class SitePercolation(NewmanZiff):
    '''A site (node) percolation experiment. This experiment computes the size of
    the giant connected component (GCC) as nodes are "occupied" within the underlying network.
    It samples the GCC at a given sequence of occupation probabilities, returning
    a time series of the growth of the GCC.

    :param g: (optional) the underlying network or generator
    :param samples: (optional) number of samples or list of sample points (defaults to 100)'''

    # Synthesised parameters
    P : Final[str] = 'epydemic.sitepercolation.pOccupied'    #: Parameter holding percolation threshold.

    # Experimental results
    GCC : Final[str] = 'epydemic.sitepercolation.gcc'        #: Result holding size of GCC.

    def __init__(self, g : Graph =None, samples : Union[int, Iterable[float]] =None):
        super(SitePercolation, self).__init__(g, samples)

    def setUp(self, params : Dict[str, Any]):
        '''Set up the process, creating the initial components data structure from the
        underlying network.

        :param params: the experimental parameters'''
        super(SitePercolation, self).setUp(params)
        N = self.network().order()
        self._unoccupied = N + 1                   # a root that can never occur
        self._components = numpy.full(N, self._unoccupied, numpy.int32)
        self._gcc = 0    # initially there are no components

    def occupy(self, nr : Node) -> int:
        '''Occupy a node, which is then joined to all adjacent occupied nodes. 
        This will update the GCC and return the size of the new component.

        This method should be overridden to collect more statistics about the network
        as components join.

        :param nr: the node
        :returns: the size of any newly-combined component'''
        g = self.network()

        # mark the node as a singleton component
        self._components[nr] = -1
        csize = 1

        # connect to all neighbouring occupied nodes
        for m in g.neighbors(nr):
            if self._components[m] != self._unoccupied:
                # neighbour is occupied, join to it
                mr = self.rootOf(m)
                
                if mr != nr:
                    csize = self.join(nr, mr)

        # update the GCC
        self._gcc = max(self._gcc, csize)

        # return the new component size
        return csize

    def sample(self, p : float) -> Dict[str, Any]:
        '''Take a sample. The default samples the size of the GCC.

        :param p: the current occupation probability
        :returns: a dict of results'''
        res = super(SitePercolation, self).sample(p)
        res[self.P] = p        
        res[self.GCC] = self._gcc        
        return res

    def percolate(self, ns : List[Node]) -> List[Dict[str, Any]]:
        '''Perform the site percolation process. This runs through the
        list of nodes, occupying them and taking samples of the occupied
        network's statistics at the requested sample points. The samples
        will be taken as the closest point above the requested probability.
        They will be recorded in the results as actually having been sampled at the
        requested point, however, so that all results are sampled with the same indices.

        :param en: the permuted list of nodes
        :returns: a list of dicts of experiment results.'''
        # take an initial sample if requested
        samples = []
        samplePoint = 0
        if self._samples[samplePoint] == 0.0:
            samples.append(self.sample(self._samples[samplePoint]))
            samplePoint += 1

        # percolate the network
        N = len(ns)
        for i in range(N):
            n = ns[i]

            # occupy the edge
            self.occupy(n)

            # take a sample if this is a sample point
            if  (i + 1) / N >= self._samples[samplePoint]:
                # we're at the closest probability after the requested sample point,
                # so build the sample
                samples.append(self.sample(self._samples[samplePoint]))

                # if we've collected all the samples we want, bail out
                samplePoint += 1
                if samplePoint > len(self._samples):
                    break
                    
        return samples

    def do(self, params : Dict[str, Any]) -> List[Dict[str, Any]]:
        '''Perform the percolation process. This passes a permuted
        list of nodes to :meth:`NewmanZiff.percolate`, which performs
        the actual operation.

        :param params: experimental parameters
        :returns: a list of dicts of experimental results'''        
        # extract and shuffle the nodes
        g = self.network()
        ns = list(g.nodes()).copy()
        numpy.random.shuffle(ns)

        # percolate the network using these nodes
        return self.percolate(ns)





# Networks process base class
#
# Copyright (C) 2017--2021 Simon Dobson
#
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from typing import Dict, List, Tuple, Any, Callable, Iterable, Union
from networkx import Graph
from epydemic import Locus, Node, Edge, Element

# There is a circular import between Process and Dynamics at the typing level
# (but not at the execution level), when providing types for dynamics() and
# setDynamics(). To deal with this we only import Dynamics in order to
# type-check Process, and not for execution.
# (See https://www.stefaanlippens.net/circular-imports-type-hints-python.html)
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from epydemic import Dynamics


# Types for handling events
EventFunction = Callable[[float, Element], None]              #: Type of event-handler functions.
EventDistribution = List[Tuple[Locus, float, EventFunction]]  #: Type of event distributions.


class Process():
    """A process that runs over a network. This is the abstract base class
    for all network processes.  It provides the essential routines to
    build, set-up, run, and extract results from a
    process. Sub-classes provide the actual behaviour by defining
    simulation events and attaching them to the network in different
    ways.

    A process definition is largely declarative, in the sense that it
    sets up event handlers and their probabilities. These are used by
    a process dynamics (a sub-class of :class:`Dynamics`) to actually
    run a simulation of the process.

    The process class includes an interface for interacting with the
    working network. The basic interface is extended and overridden in
    different sub-classes that interact with the network in different
    ways.

    """

    # defaults
    DEFAULT_MAX_TIME = 20000      #: Default maximum simulation time.

    def __init__(self):
        super().__init__()

        # set the default maximum time, which persists across runs of the process
        self._maxTime = self.DEFAULT_MAX_TIME

        # reset this instance
        #self.reset()


    # ---------- Setup and initialisation ----------

    def reset(self):
        """Reset the process ready to be built. This resets all the internal
        process state variables. The default does nothng.

        """
        pass

    def build(self, params : Dict[str, Any]):
        """Build the process model. This should be overridden by sub-classes,
        and should create the various elements of the model.

        :param params: the model parameters

        """
        pass

    def setUp(self, params : Dict[str, Any]):
        """Set up the network under the given dynamics. The default does
        nothing: sub-classes should override it to initialise node states
        or establish other properties ready for the experiment.

        :param params: the simulation parameters

        """
        pass

    def tearDown(self):
        """Tear down any structures built for this run of the process. The
        default does nothing.

        """
        pass


    # ---------- State access and update ----------

    def setNetwork(self, g : Graph):
        """Set the network the process is running over.

        :param g: the network

        """
        self._dynamics.setNetwork(g)

    def network(self) -> Graph:
        """Return the network the process is running over.

        :returns: the network

        """
        return self._dynamics.network()

    def setDynamics(self, d : 'Dynamics'):
        '''Set the instance of :class:`Dynamics` that runs the process.

        :param d: the dynamics'''
        self._dynamics = d

    def dynamics(self) -> 'Dynamics':
        '''Return the instance of :class:`Dynamics` running this process.

        :returns: the dynamics

        '''
        return self._dynamics

    def setMaximumTime(self, t : float):
        """Set the maximum default simulation time. The default is given by
        :attr:`DEFAULT_MAX_TIME`.  This is used by
        :meth:`atEquilibrium` as the default way to determine
        equilibrium.

        The maximum time may be slightly exceeded due to the ways in
        which events are drawn.

        Setting the maximum time persists across runs of the process,
        and isn't reset to its default by calling :meth:`reset`.

        :param t: the maximum simulation time

        """
        self._maxTime = t

    def maximumTime(self) -> float:
        """Return the maximum assumed simulation time.

        :returns: the maximum simulation time"""
        return self._maxTime


    # ---------- Termination and results ----------

    def atEquilibrium(self, t : float) -> bool:
        """Test whether the process is an equilibrium. The default simply
        checks whether the simulation time exceeds the maximum given
        by :meth:`setMaximumTime`, which defaults to an arbitrary
        value given by :attr:`DEFAULT_MAX_TIME`.  Sub-classes may
        override this to provide a more sensible definition: in many
        cases it makes sense to keep calling this method as part of
        any more advanced test, so as to cut-off runaway processes.

        :param t: the current simulation time
        :returns: True if the proceess is now at equilibrium

        """
        return (t >= self.maximumTime())

    def results(self) -> Dict[str, Any]:
       """Create and return an empty dict to be filled with experimental
       results.  Sub-classes should extend this method to add results
       to the dict.

       :returns: an empty dict for experimental results

       """
       return dict()


    # ---------- Accessing and evolving the network ----------

    def addNode(self, n : Node, **kwds):
        """Add a node to the working network. Any keyword arguments added as node attributes

        :param n: the new node
        :param kwds: (optional) node attributes"""
        self.network().add_node(n, **kwds)

    def addNodesFrom(self, ns : Iterable[Node], **kwds):
        """Add all the nodes in the given iterable to the working network. Any
        keyword arguments are added as node attributes. This works by
        calling :meth:`addNode` for each element of the iterable.

        :param ns: an iterable collection of nodes
        :param kwds: (optional) node attributes

        """
        for n in ns:
            self.addNode(n, **kwds)

    def removeNode(self, n : Node):
        """Remove a node from the working network.

        :param n: the node"""
        self.network().remove_node(n)

    def removeNodesFrom(self, ns : Iterable[Node]):
        """Remove all the nodes in the given iterable from the working
        network. This works by iteratively calling :meth:`removeNode`
        for each element of the iterable collection.

        :param ns: an iterable collection of nodes

        """
        for n in ns:
            self.removeNode(n)

    def addEdge(self, n : Node, m : Node, **kwds):
        """Add an edge between nodes. Any keyword arguments are added as edge
        attributes.  If the endpoint nodes do not already exist then
        an exception is raised.

        :param n: the start node
        :param m: the end node
        :param kwds: (optional) edge attributes

        """
        g = self.network()
        if n not in g:
            raise Exception('No node {n} in network'.format(n = n))
        if m not in g:
            raise Exception('No node {n} in network'.format(n = m))
        g.add_edge(n, m, **kwds)

    def addEdgesFrom(self, es : Iterable[Edge], **kwds):
        """Add all the edges in the given iterable to the working network. Any
        keyword arguments are added as node attributes. This works by
        calling :meth:`addEdge` for each element of the iterable.

        :param es: an iterable collection of edges
        :param kwds: (optional) node attributes

        """
        for e in es:
            (n, m) = e
            self.addEdge(n, m, **kwds)

    def removeEdge(self, n : Node, m : Node):
        """Remove an edge from the working network.

        :param n: the start node
        :param m: the end node"""
        self.network().remove_edge(n, m)

    def removeEdgesFrom(self, es : Iterable[Edge]):
        """Remove all the edges in the given iterable collection from the
        working network. This works by iteratively calling
        :meth:`removeEdge` for each edge in the iterable.

        :param es: an iterable collection of edges

        """
        for e in es:
            self.removeEdge(*e)


    # ---------- Probabilistic and posted events ----------
    # These are helper methods that delegate to the dynamics

    def addLocus(self, n : str, l : Locus =None) -> Locus:
        """Add a named locus.

        :param n: the locus name
        :param l: the locus (defaults to a simple set-based locus)
        :returns: the locus"""
        return self._dynamics.addLocus(self, n, l)

    def loci(self) -> Dict[str, Locus]:
        '''Return the names of the loci that this process added.

        :returns: a dict from names to loci'''
        return self._dynamics.lociForProcess(self)

    def locus(self, n : str) -> Locus:
        '''Return the named locus.

        :param n: the locus name
        :returns: the locus'''
        return self.loci()[n]

    def addEventPerElement(self, l : Union[str, Locus], p : float, ef : EventFunction):
        """Add a probabilistic event at a locus, occurring with a particular
        (fixed) probability for each element of the locus, and calling
        the :term:`event function` when it is selected. The locus may
        be a :class:`Locus` object or a string, which is taken to be a
        locus of this process.

        Unlike fixed probability events added by
        :meth:`addFixedRateEvent`, a per-element event happens with
        the same probability to each element in the locus. This leads
        to larger loci giving rise to a higher rate of events.

        :param l: the locus or locus name
        :param p: the event probability
        :param ef: the event function

        """
        self._dynamics.addEventPerElement(self, l, p, ef)

    def addFixedRateEvent(self, l : Union[str, Locus], p : float, ef : EventFunction):
        """Add a probabilistic event at a locus, occurring with a particular
        (fixed) probability, and calling the :term:`event function`
        when it is selected. The locus may be a :class:`Locus` object
        or a string, which is taken to be the name of a locus of this
        process.

        Unlike fixed rate events added by :meth:`addEventPerElement`,
        a fixed probability event happens with the same probability
        regardless of how many elements are in the locus.

        :param l: the locus or locus name
        :param p: the event probability
        :param ef: the event function

        """
        self._dynamics.addFixedRateEvent(self, l, p, ef)

    def postEvent(self, t : float, e : Any, ef : EventFunction):
        """Post an event that calls the :term:`event function` at time t.

        :param t: the current time
        :param e: the element (node or edge) on which the event occurs
        :param ef: the event function"""
        self._dynamics.postEvent(t, e, ef)

    def postRepeatingEvent(self, t : float, dt : float, e : Any, ef : EventFunction):
        """Post an event that starts at time t and re-occurs at interval dt.

        :param t: the start time
        :param dt: the interval
        :param e: the element (node or edge) on which the event occurs
        :param ef: the element function

        """
        self._dynamics.postRepeatingEvent(t, dt, e, ef)



# Standard generators
#
# Copyright (C) 2017--2020 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import NetworkGenerator
from networkx import Graph, fast_gnp_random_graph, barabasi_albert_graph
import sys
if sys.version_info >= (3, 8):
    from typing import Any, Dict, Optional, Final
else:
    # backport compatibility with older typing
    from typing import Any, Dict, Optional
    from typing_extensions import Final

class FixedNetwork(NetworkGenerator):
    '''A network generator that always returns a copy of the same network.

    :param g: the prototype network
    :param limit: (optional) maximum number of identical instances to generate'''

    def __init__(self, g : Graph, limit : Optional[int] =None):
        super(FixedNetwork, self).__init__(limit=limit)
        self._graphPrototype : Graph = g
    
    def _generate(self, params : Dict[str, Any]) -> Graph:
        '''Return a copy of the prototype network.

        :param params: experimental parameters (ignored)
        :returns: a network instance'''
        return self._graphPrototype.copy()


class ERNetwork(NetworkGenerator):
    '''Generate Erdos-Renyi (ER) networks from a given order (:attr:`N`) and one of an edge occupation
    probability (:attr:`PHI`) or a mean degree (:attr:`KMEAN`). These parameters are taken from the
    experimental parameters.

    An ER network has nodes with Poisson-distributed independent degrees. The construction process
    can be thought of as taking a set of :math:`N` nodes and then adding an edge between every pair
    with independent probability :math:`\\phi`. This process gives a discrete normal (Poisson) distribution
    of the node degrees with mean degree :math:`\\langle k \\rangle = N \\phi`. The node degrees will
    be uncorrelated.

    The actual construction of ER networks uses the `networkx.fast_gnp_random_graph()` function.

    :param params: (optional) experiment parameters
    :param limit: (optional) meximum  number of instances to generate'''

    # Experimental parameters
    N : Final[str] = 'epydemic.generators.ERNetwork.N'         #: Experimental parameter for the size (order) of the network.
    PHI : Final[str] = 'epydemic.generators.ERNetwork.PHI'     #: Experimental parameter for the occupation probability of edges.
    KMEAN : Final[str] = 'epydemic.generators.ERNetwork.KMEAN' #: Experimental parameter for the mean degree of the network.

    def __init__(self, params : Dict[str, Any] =None, limit : Optional[int] =None):
        super(ERNetwork, self).__init__(params, limit)

    def _generate(self, params : Dict[str, Any]) -> Graph:
        '''Generate an ER network from an order (represented by the parameter :attr:`N`)
        and one of an edge occupation probability (:attr:`PHI`) or mean degree (:attr:`KMEAN`).

        :param params: experimental parameters
        :returns: the ER network'''

        # extract the parameters
        N = params[self.N]
        if self.PHI in params:
            phi = params[self.PHI]
        elif self.KMEAN in params:
            kmean = params[self.KMEAN]
            phi = (kmean + 0.0) / N
        else:
            raise AttributeError('Need one of occupation probability or mean degree')

        # build the network
        g = fast_gnp_random_graph(N, phi)
        return g


class BANetwork(NetworkGenerator):
    '''Generate Barabasi-Albert (BA) networks from a given order (:attr:`N`) and rate of attachment
    (:attr:`M`) taken from the experimental parameters.

    A BA network has node degrees distributed according to a powerlaw distribution. The construction process
    can be thought of as taking an initial set of :math:`M` nodes and then adding additional nodes
    one at a time, adding adges between the new node and :math:`M` other nodes chosen at random. This
    continues until the network has :math:`N` nodes. This process favours attachment to nodes that are 
    in the network early, leading to "hubs" with very high degree. The node degrees will
    be uncorrelated.

    The actual construction of BA networks uses the `networkx.barabasi_albert_graph()` function.

    :param params: (optional) experiment parameters
    :param limit: (optional) meximum  number of instances to generate'''

    # Experimental parameters
    N : Final[str] = 'epydemic.generators.BANetwork.N'         #: Experimental parameter for the size (order) of the network.
    M : Final[str] = 'epydemic.generators.BANetwork.M'         #: Experimental parameter for the number of edges added per node.

    def __init__(self, params : Dict[str, Any] =None, limit : Optional[int] =None):
        super(BANetwork, self).__init__(params, limit)

    def _generate(self, params : Dict[str, Any]) -> Graph:
        '''Generate a BA network from an order (represented by the parameter :attr:`N`)
        and attachment rate (:attr:`M`).

        :param params: experimental parameters
        :returns: the ER network'''

        # extract the parameters
        N = params[self.N]
        M = params[self.M]

        # build the network
        g = barabasi_albert_graph(N, M)
        return g
